#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import logging
import tempfile
import re
import warnings

from pydantic import BaseSettings

PATTERN_MAC_STRIPPED = r"^[a-f0-9]{12}$"
REGEX_MAC_STRIPPED = re.compile(PATTERN_MAC_STRIPPED)

PATTERN_VALID_HEX = r"^[a-f0-9]{4,}$"
REGEX_VALID_HEX = re.compile(PATTERN_VALID_HEX)

PATTERN_RPI_SERIAL_HEX = r"^[a-f0-9]{8}$"
REGEX_RPI_SERIAL_HEX = re.compile(PATTERN_RPI_SERIAL_HEX, re.I)

PATTERN_WIFI_HOST = r"^192\.168\.205\.\d+$"
REGEX_WIFI_HOST = re.compile(PATTERN_WIFI_HOST)

WIFI_GW_FAKE_MAC = "baaaaaadc0de"


class Settings(BaseSettings):
    edge_domain: str = "example.com"


CONFIGURATION = Settings()

if CONFIGURATION.edge_domain == "example.com":
    warnings.warn(f"Edge Domain is {CONFIGURATION.edge_domain!r}!")


def valid_colon_separated_mac_addr_or_bust(mac_addr):
    """
    Validate that value contains a valid colon separated MAC address.

    Args:
        mac_addr (str): alleged MAC address

    Returns:
        tuple: stripped address, portions, vendor prefix

    >>> valid_colon_separated_mac_addr_or_bust("")
    Traceback (most recent call last):
        ...
    ValueError: invalid literal for int() with base 16: ''
    >>> valid_colon_separated_mac_addr_or_bust("00:30:18:09:c6:b2")
    ('00301809c6b2', [0, 48, 24, 9, 198, 178], '00:30:18')
    >>> valid_colon_separated_mac_addr_or_bust("00:30:18:09:c6:b2")
    ('00301809c6b2', [0, 48, 24, 9, 198, 178], '00:30:18')
    >>> valid_colon_separated_mac_addr_or_bust("00-30-18-09-c6-b2")
    Traceback (most recent call last):
        ...
    ValueError: invalid literal for int() with base 16: '00-30-18-09-c6-b2'
    >>> valid_colon_separated_mac_addr_or_bust("ba:aa:aa:ad:c0:de")
    ('baaaaaadc0de', [186, 170, 170, 173, 192, 222], 'ba:aa:aa')
    """
    mac_addr = mac_addr.lower()
    mac_portions = [int(x, 16) for x in mac_addr.split(":")]

    mac_stripped = "".join(["{:02x}".format(x) for x in mac_portions])
    vendor_prefix = ":".join(["{:02x}".format(x) for x in mac_portions[:3]])

    assert REGEX_MAC_STRIPPED.match(mac_stripped) is not None

    return mac_stripped, mac_portions, vendor_prefix


class RemoteBerryDataMismatch(ValueError):
    pass


class RemoteBerryBadData(ValueError):
    pass


class RemoteBerryIdentity:
    """
    Raspberry Pi device identity model class.
    """

    def __init__(self, in_data, *args, **kwargs):
        self.log = logging.getLogger(__name__)
        self.configuration_data_path = kwargs.get(
            "configuration_data_path", tempfile.gettempdir()
        )
        self.data = dict(identity=None, rpi_serial=None, gw_mac=None)
        self._in_data = in_data

        try:
            self._parse()
        except Exception as exc:
            self.log.warning(exc)

    def _parse(self):
        try:
            self.data["identity"] = self._in_data["identity"]
        except KeyError:
            pass

        try:
            self.data["rpi_serial"] = self._in_data["otp"]["rpi_serial"]
            assert REGEX_VALID_HEX.match(self.data["rpi_serial"]) is not None
        except (KeyError, AssertionError, TypeError):
            pass

        try:
            self.data["gw_mac"], _, _ = valid_colon_separated_mac_addr_or_bust(
                self._in_data["gateway"]["mac_addr"]
            )
            assert REGEX_VALID_HEX.match(self.data["gw_mac"]) is not None
        except (KeyError, AssertionError, TypeError):
            pass

        if self.data["identity"]:
            value = self.data["identity"]["gw_mac"]

            if not REGEX_MAC_STRIPPED.match(value):
                self.log.warning("Tainted: mac_stripped={!r}".format(value))

        if self.data["identity"] and not self.data["rpi_serial"]:
            try:
                self.data["rpi_serial"] = self._in_data["identity"][
                    "rpi_serial"
                ]
            except KeyError:
                pass

        if self.data["gw_mac"] and self.data["identity"]:
            if self.data["gw_mac"] != self.data["identity"].get("gw_mac"):
                self.log.warning(
                    "Tainted: gw_mac={!r} identity/gw_mac={!r}".format(
                        self.data["gw_mac"], self.data["identity"]["gw_mac"]
                    )
                )

    def __str__(self):
        portions = [
            "<{:s}".format(self.__class__.__name__),
            (not self.sufficient and " (INSUFFICIENT)" or ""),
        ]

        try:
            portions.append(" {:s}".format(self.inventory_hostname))
        except Exception:
            pass

        if self.rpi_serial:
            portions.append(" rpi_serial={:s}".format(self.rpi_serial))

        if self.gw_mac:
            portions.append(" gw_mac={:s}".format(self.gw_mac))

        try:
            portions.append(
                " serial={!r}".format(self.identity["serial_number"])
            )
        except Exception:
            pass

        portions.append(">")

        return "".join(portions)

    @property
    def sufficient(self):
        return {self.rpi_serial, self.gw_mac} != {None}

    @property
    def data_path(self):
        if self.gw_mac:
            return os.path.join(self.configuration_data_path, self.gw_mac)

        if not self.data["rpi_serial"]:
            raise RemoteBerryBadData("rpi_serial")

        return self.rpi_data_path

    @property
    def rpi_data_path(self):
        if self.rpi_serial:
            return os.path.join(self.configuration_data_path, self.rpi_serial)

        return None

    @property
    def storage_key(self):
        if self.rpi_serial:
            return self.rpi_serial

        return None

    @property
    def inventory_hostname(self):
        if not self.identity.get("inventory_hostname"):
            raise RemoteBerryBadData("inventory_hostname")

        if self.identity.get("inventory_hostname").strip().startswith("{{"):
            raise RemoteBerryBadData("inventory_hostname")

        return self.identity["inventory_hostname"]

    @property
    def inventory_hostname_for_storage(self):
        if self.gw_mac == WIFI_GW_FAKE_MAC:
            return f"{self.rpi_serial}.{CONFIGURATION.edge_domain}"

        if self.inventory_hostname.startswith("192.168."):
            return f"{self.rpi_serial}.{CONFIGURATION.edge_domain}"

        return self.inventory_hostname

    @property
    def gw_mac(self):
        return self.data["gw_mac"]

    @property
    def identity(self):
        if not self.data["identity"]:
            raise RemoteBerryBadData("identity")

        return self.data["identity"]

    @property
    def rpi_serial(self):
        return self.data["rpi_serial"]


if __name__ == "__main__":
    import doctest

    (FAILED, SUCCEEDED) = doctest.testmod()
    print("[doctest] SUCCEEDED/FAILED: {:d}/{:d}".format(SUCCEEDED, FAILED))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import copy
import logging

import pendulum
import orjson

from coshed_model.ansible_remote_identity import RemoteBerryIdentity
from coshed_model.ansible_remote_identity import REGEX_WIFI_HOST

COUCHDB_INTERNAL_KEYS = ("_id", "_rev")

#: key of key/value pairs holding dataset
DATA_KEYS = (
    "tags",
    "host",
    "dump",
    "location",
    "ping",
    "snapshot",
    "ring",
    "time_out",
    "taints",
    "flags",
)

#: key of the key/value pairs keeping track on when datasets WERE updated
PORTIONS_DT_KEY = "portions_dt"

#: key of the key/value pairs keeping track on when datasets SHOULD be updated
PORTIONS_NEXT_DT_KEY = "portions_next_dt"

#: current version of the data model specification
SPECIFICATION_VERSION = 10

#: automatic refreshing of datasets thresholds
AUTOMATIC_REFRESH_THRESHOLDS = {
    "_default": pendulum.duration(days=1),
    "_empty_value": pendulum.duration(hours=9),
    "ping": pendulum.duration(minutes=9),
    "location": pendulum.duration(days=7),
    "dump": pendulum.duration(minutes=30),
    "snapshot": pendulum.duration(minutes=30),
}


def mangled_data_portion(key, data):
    log = logging.getLogger(__name__)
    log.debug("<<< {!s}: {!r}".format(key, data))

    if data is None:
        return None

    del_keys = set(["dt"]) | set(COUCHDB_INTERNAL_KEYS)

    if key == "location":
        del_keys |= {"objectId", "internalID", "imageUrl", "contacts"}

        for data_key in data.keys():
            if not data[data_key]:
                del_keys.add(data_key)
    elif key == "host" and data.get("fishy") is None:
        data["is_fishy"] = False
        try:
            if REGEX_WIFI_HOST.match(data["inventory_hostname"]):
                data["is_fishy"] = True
        except Exception as exc:
            log.error(exc)

    if isinstance(data, dict):
        for del_key in del_keys:
            try:
                del data[del_key]
            except KeyError:
                pass

    log.debug(">>> {!s}: {!r}".format(key, data))

    return data


class AHMDocument(dict):
    """
    Ansible Host Metadata Document.

    >>> a = AHMDocument()
    >>> b = AHMDocument()
    >>> a == b
    True
    >>> a["x"] = 123
    >>> a.dirty
    True
    >>> not a.has_changed_portion({"something_else"})
    True
    >>> a["tags"] = ["a", "b", "c"]
    >>> a.dirrty_filthy_nasty
    {'tags'}
    >>> not a.has_changed_portion({"tags"})
    False
    >>> json_bytes = orjson.dumps(a)
    >>> len(json_bytes)
    204
    >>> set(b.keys()) == {'portions_dt', 'portions_next_dt'}
    True
    >>> len(b.keys())
    2
    >>> b.remangle()
    >>> len(b.keys()) > 2
    True
    """

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        self.log = logging.getLogger(__name__)

        for cik in COUCHDB_INTERNAL_KEYS:
            try:
                del self[cik]
            except KeyError:
                pass

        self.dirrty_filthy_nasty = set()
        self.read_dt = copy.copy(self.dt)

        for portions_dk in (PORTIONS_DT_KEY, PORTIONS_NEXT_DT_KEY):
            try:
                self[portions_dk]
            except KeyError:
                self[portions_dk] = dict()

            if not isinstance(self[portions_dk], dict):
                self[portions_dk] = dict()

    def next_refresh_threshold(self, key):
        """
        Determine the threshold value for automatic refresh of ``key``.

        Args:
            key (str): data portion key

        Returns:
            datetime.timedelta: auto refresh threshold
        """
        if self.get(key) is None:
            return AUTOMATIC_REFRESH_THRESHOLDS["_empty_value"]
        else:
            try:
                return AUTOMATIC_REFRESH_THRESHOLDS[key]
            except KeyError:
                pass

        return AUTOMATIC_REFRESH_THRESHOLDS["_default"]

    def touch(self, portion_key=None):
        """
        Mark metadata document as **changed** storing current timestamp as
        change date.

        If ``portion_key`` is provided specifically also mark the referenced
        data portion as changed.

        Args:
            portion_key (str, optional): data portion key. Defaults to None.

        """
        self.log.debug("You can't touch this")
        u_now = pendulum.now(pendulum.UTC)
        self["dt"] = u_now.to_rfc3339_string()
        self["version"] = SPECIFICATION_VERSION

        if portion_key:
            self.dirrty_filthy_nasty.add(portion_key)
            self[PORTIONS_DT_KEY][portion_key] = self["dt"]
            threshold = self.next_refresh_threshold(portion_key)
            next_dt = u_now + threshold
            self[PORTIONS_NEXT_DT_KEY][
                portion_key
            ] = next_dt.to_rfc3339_string()

    @property
    def dirty(self):
        """
        Return indicator if metadata should be persisted.

        Returns:
            bool: ``True`` if document has changed
        """
        return self.read_dt != self.dt

    @property
    def dt(self):
        """
        Return *latest changed* datetime object.

        Returns:
            datetime.datetime: datetime of latest change. Falls back to \
                current datetime
        """
        try:
            return pendulum.parse(self["dt"])
        except KeyError:
            return pendulum.now(pendulum.UTC)

    @property
    def version(self):
        """
        Return document version

        Returns:
            int: document version. Falls back to ``0``
        """
        try:
            return self["version"]
        except KeyError:
            return 0

    @property
    def identity(self):
        """
        Return remote berry identity object

        Returns:
            RemoteBerryIdentity: identity object
        """
        return RemoteBerryIdentity(dict(identity=self["host"]))

    @property
    def equipment_id(self):
        """
        Return AIN equipment ID. Falls back to ``None``.

        Returns:
            str: AIN ID
        """
        try:
            return self["host"]["equipment_id"]
        except Exception:
            pass

        return None

    @property
    def item_id(self):
        """
        Return Thing ID.  Falls back to ``None``.

        Returns:
            str: Thing ID
        """
        try:
            return self["host"]["item_id"]
        except Exception:
            pass

        return None

    @property
    def serial_number(self):
        """
        Return Serial Number. Falls back to ``None``.

        Returns:
            str: serial number
        """
        try:
            return self["host"]["serial_number"]
        except Exception:
            pass

        return None

    @property
    def inventory_hostname(self):
        """
        Return ansible inventory hostname. Falls back to ``None``.

        Returns:
            str: ansible inventory hostname
        """
        try:
            return self["host"]["inventory_hostname"]
        except Exception:
            pass

        return None

    @property
    def is_fishy(self):
        """
        Return indicator if current metadata indicates a *fishy*
        host. This means that the host is apparently not connected
        by a genubox device.

        Returns:
            bool: Indicator if host is to be considered *fishy*
        """
        try:
            return self["host"]["is_fishy"]
        except Exception:
            pass

        return False

    def has_changed_portion(self, interest):
        return self.dirrty_filthy_nasty & interest

    def next_refresh(self, key):
        """
        Determine the next moment in time at which ``key`` may be refreshed
        automatically.

        Args:
            key (str): data portion key

        Returns:
            datetime.datetime: next auto refresh
        """
        threshold = self.next_refresh_threshold(key)
        next_dt = self.read_dt + threshold

        try:
            next_dt = pendulum.parse(self[PORTIONS_NEXT_DT_KEY][key])
        except KeyError:
            self[PORTIONS_NEXT_DT_KEY][key] = next_dt.to_rfc3339_string()
        except Exception:
            pass

        return next_dt

    def refreshable_portion(self, key):
        """
        Return indicator if data portion ``key`` may be refreshed now.

        Args:
            key (str): data portion key

        Returns:
            bool: ``True`` if data portion may be refreshed
        """
        return pendulum.now() >= self.next_refresh(key)

    def __setitem__(self, key, value):
        old_value = self.get(key)

        if key in DATA_KEYS:
            value = mangled_data_portion(key, copy.deepcopy(value))

        super().__setitem__(key, value)

        if key in DATA_KEYS and old_value != value:
            self.touch(key)

    def remangle(self):
        """
        (Re-)Apply mangling to all data portion key/value pairs.
        """
        for key in DATA_KEYS:
            self[key] = self.get(key)

    def __str__(self):
        if self.dirty:
            dirty_indicator = " DIRTY ({:s})".format(
                ", ".join(sorted(self.dirrty_filthy_nasty))
            )
        else:
            dirty_indicator = ""

        return "<{klass} {version}> {dt}{dirty_indicator}".format(
            klass=self.__class__.__name__,
            version=self.version,
            dt=self.dt,
            dirty_indicator=dirty_indicator,
        )

    def __eq__(self, other):
        for key in DATA_KEYS:
            if self.get(key) != other.get(key):
                return False

        return True

    def json(self):
        return orjson.dumps(self)


if __name__ == "__main__":
    import doctest

    (FAILED, SUCCEEDED) = doctest.testmod()
    print("[doctest] SUCCEEDED/FAILED: {:d}/{:d}".format(SUCCEEDED, FAILED))
